resnet18_url = 'https://www.dropbox.com/s/ln4i4547jvn5wr8/resnet18.pth?dl=1'
resnet34_url = 'https://www.dropbox.com/s/2mxj66dqtfqss2c/resnet34.pth?dl=1'
resnet50_url = 'https://www.dropbox.com/s/hinaadj4kj1rsl0/resnet50.pth?dl=1'
resnet101_url = 'https://www.dropbox.com/s/11u40th2ygsz9u0/resnet101.pth?dl=1'
resnet152_url = 'https://www.dropbox.com/s/9babg74hcgx1w9l/resnet152.pth?dl=1'
elephant_skin_url = 'https://www.dropbox.com/s/xak3qtgy426nwig/elephant_skin%281%29.pth?dl=1'
snake_skin_url = 'https://www.dropbox.com/s/98nmx8gcsce49zr/snake_skin.pth?dl=1'
lenet_mnist_url = 'https://www.dropbox.com/s/0ylrj6a8gx1iiuw/lenet_mnist.pth?dl=1'
alexnet_url = 'https://www.dropbox.com/s/hogfy6yap8an5x9/alexnet.pth?dl=1'
alexnet_mnist_url = 'https://www.dropbox.com/s/u123zs9t6rcco18/alexnet_mnist.pth?dl=1'

# This is a test dataset which is very small in size and 
# helps ensures things are working as they are supposed to
tdset = 'https://www.dropbox.com/s/79cnd89nxjm60u6/testdataset.zip?dl=1'
